/*

    ThermostatQA - test framework for Thermostat Monitoring Tool

    Copyright 2013 Red Hat, Inc.

This file is part of ThermostatQA

ThermostatQA is distributed under the GNU General Public License,
version 2 or any later version (with a special exception described
below, commonly known as the "Classpath Exception").

A copy of GNU General Public License (GPL) is included in this
distribution, in the file COPYING.

Linking ThermostatQA code with other modules is making a combined work
based on ThermostatQA.  Thus, the terms and conditions of the GPL
cover the whole combination.

As a special exception, the copyright holders of ThermostatQA give you
permission to link this code with independent modules to produce an
executable, regardless of the license terms of these independent
modules, and to copy and distribute the resulting executable under
terms of your choice, provided that you also meet, for each linked
independent module, the terms and conditions of the license of that
module.  An independent module is a module which is not derived from
or based on ThermostatQA code.  If you modify ThermostatQA, you may
extend this exception to your version of the software, but you are
not obligated to do so.  If you do not wish to do so, delete this
exception statement from your version.
*/

package org.thermostat.qa.reporter;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;



/**
 * History page(s) generator. It uses HTML template hist.html
 * stored in "templates" subdirectory.
 *
 * @author Pavel Tisnovsky <ptisnovs@redhat.com>
 */
public class HistoryPagesGenerator implements CommonGenerator
{

    private static boolean debug = true;

    public static void generate(Map<String, Map<String, List<String>>> testResults, CommandLineParameters params)
    {
        createHtmlPageFromTemplate(testResults, params, 10, TestType.FAILED);
        createHtmlPageFromTemplate(testResults, params, 20, TestType.FAILED);
        createHtmlPageFromTemplate(testResults, params, 30, TestType.FAILED);
        createHtmlPageFromTemplate(testResults, params, -1, TestType.ALL);
        createHtmlPageFromTemplate(testResults, params, -1, TestType.FAILED);
    }

    private static void createHtmlPageFromTemplate(Map<String, Map<String, List<String>>> testResults,
                    CommandLineParameters params, int testCount, TestType testType)
    {
        List<String> template = FileUtils.readTextFile(params.getTemplateDir() + "/hist.html");
        List<String> out = new LinkedList<String>();

        // list of dates
        List<String> dateList = getDateList(testResults, testCount, testType);

        if (debug)
        {
            printDateList(dateList);
        }

        // list of all tests
        Set<String> testList = getTestList(testResults, dateList);

        if (debug)
        {
            printTestList(testList);
        }

        // iterate through whole template
        for (String templateLine : template) {
            // replace text in template where needed
            if (templateLine.contains("${RESULTS}"))
            {
                String testCountMessage = generateTitleMessage(testCount, testType);
                templateLine = templateLine.replace("${RESULTS}", testCountMessage);
                out.add(templateLine);
            }
            else if (templateLine.contains("${TABLE_DATA}"))
            {
                renderTableData(out, testResults, dateList, testList);
            }
            else
            {
                out.add(templateLine);
            }
        }

        // write list of string to a file with given name
        String outName = createOutputFileName(testCount, testType);
        FileUtils.writeTextFile(params.getReportDir() + "/" + outName + ".html", out);
    }
    
    private static void printDateList(List<String> dateList) {
        System.out.println("Date list:");
        printStringCollection(dateList);
    }

    private static void printTestList(Set<String> testList) {
        System.out.println("Test list:");
        printStringCollection(testList);
    }

    /**
     * @param dateList
     */
    private static void printStringCollection(Collection<String> dateList) {
        for (String date : dateList) {
            System.out.println("    " + date);
        }
    }

    private static List<String> getDateList(Map<String, Map<String, List<String>>> testResults, int testCount, TestType testType) {
        List<String> dates = new LinkedList<String>();
        for (Map.Entry<String, Map<String, List<String>>> oneResult : testResults.entrySet()) {
            dates.add(oneResult.getKey());
        }
        if (testCount >= 0) {
            int fromIndex = Math.max(0, dates.size() - testCount);
            dates = dates.subList(fromIndex, dates.size());
        }
        return dates;
    }

    private static Set<String> getTestList(Map<String, Map<String, List<String>>> testResults, List<String> dateList) {
        Set<String> failedTests = new TreeSet<String>();

        for (String date : dateList)
        {
            for (Map.Entry<String, List<String>> test : testResults.get(date).entrySet())
            {
                failedTests.addAll(grepTests(test.getValue(), "FAILED: "));
            }
        }
        if (debug)
        {
            System.out.println("# of failed tests: " + failedTests.size());
        }

        for (int i = 0; i < dateList.size() - 1; i++)
        {
            String date1 = dateList.get(i);
            String date2 = dateList.get(i + 1);
            Set<String> passedTests1 = new TreeSet<String>();
            Set<String> passedTests2 = new TreeSet<String>();
            
            for (Map.Entry<String, List<String>> test : testResults.get(date1).entrySet())
            {
                passedTests1.addAll(grepTests(test.getValue(), "PASSED: "));
            }
            for (Map.Entry<String, List<String>> test : testResults.get(date2).entrySet())
            {
                passedTests2.addAll(grepTests(test.getValue(), "PASSED: "));
            }

            Set<String> diff1 = new TreeSet<String>();
            Set<String> diff2 = new TreeSet<String>();
            diff1.addAll(passedTests1);
            diff2.addAll(passedTests2);
            diff1.removeAll(passedTests2);
            diff2.removeAll(passedTests1);
            failedTests.addAll(diff1);
            failedTests.addAll(diff2);
        }
        return failedTests;
    }

    /**
     * @param failedTests
     * @param log
     */
    private static Set<String> grepTests(List<String> log, String prefix) {
        Set<String> result = new TreeSet<String>();
        for (String line : log)
        {
            if (line.startsWith(prefix))
            {
                String testName = line.substring(prefix.length());
                if (testName.indexOf(' ') >= 0)
                {
                    testName = testName.substring(0, testName.indexOf(' '));
                }
                if (testName.indexOf(':') >= 0)
                {
                    testName = testName.substring(0, testName.indexOf(':'));
                }
                result.add(testName);
            }
        }
        return result;
    }

    /**
     * @param testCount
     * @param testType
     * @return
     */
    private static String generateTitleMessage(int testCount, TestType testType) {
        return testCount == -1 ? testType ==  TestType.ALL ? "all results" : "failed tests" : "last " + testCount + " results";
    }
    /**
     * @param testCount
     * @return
     */
    private static String createOutputFileName(int testCount, TestType testType) {
        String s = testCount == -1 ? testType == TestType.ALL ? "all_tests" : "failed_tests" : "" + testCount;
        return "hist_" + s;
    }

    private static void renderTableData(List<String> out, Map<String, Map<String, List<String>>> testResults, List<String> dateList, Set<String> testList) {
        printFirstTableLine(out, dateList);
        printRestOfTable(out, testResults, dateList, testList);
    }

    /**
     * @param out
     * @param dateList
     */
    private static void printFirstTableLine(List<String> out, List<String> dateList) {
        out.add("<tr><td class='group-id'>Test name/Date:</td>");
        for (String date : dateList)
        {
            String d1 = date.substring(0, 4);
            String d2 = date.substring(5);
            out.add("<td>" + d1 + "<br />" + d2 + "</td>");
        }
        out.add("</tr>");
    }

    private static void printRestOfTable(List<String> out, Map<String, Map<String, List<String>>> testResults, List<String> dateList, Set<String> testList) {
        for (String testName : testList)
        {
            String javaDocHref = generateHrefToJavaDoc(testName);
            out.add("<tr><td><a href='javadoc/" + javaDocHref + "'>" + javaDocHref + " _ " + testName + "</a></td>");
            for (String date : dateList)
            {
                TestStatus testStatus = getTestStatus(testName, date, testResults);
                String cssClass = getCssClass(testStatus);
                String cellText = getCellText(testStatus);
                if (testStatus == TestStatus.NOT_FOUND)
                {
                    out.add("<td class='" + cssClass + "'>" + cellText + "</td>");
                }
                else
                {
                    String href = "log_" + date + ".html";
                    out.add("<td class='" + cssClass + "'><a href='" + href + "#"+ testName + "'>" + cellText + "</a></td>");
                }
            }
            out.add("</tr>");
        }
    }

    /**
     * @param testStatus
     * @return
     */
    private static String getCssClass(TestStatus testStatus) {
        String cssClass = testStatus == TestStatus.PASSED ? "passed-header" :
                          testStatus == TestStatus.FAILED ? "failed-header" :
                          testStatus == TestStatus.IGNORED ? "ignored-header" :
                          "empty";
        return cssClass;
    }

    /**
     * @param testStatus
     * @return
     */
    private static String getCellText(TestStatus testStatus) {
        String cellText = testStatus == TestStatus.PASSED ? "ok" :
                          testStatus == TestStatus.FAILED ? "fail" :
                          testStatus == TestStatus.IGNORED ? "x" :
                          "&nbsp;";
        return cellText;
    }

    /**
     * @param testName
     * @return
     */
    private static String generateHrefToJavaDoc(String testName)
    {
        String javaDocHref = testName.replace('.', '/');
        javaDocHref = javaDocHref.substring(0, javaDocHref.lastIndexOf('/'))
                    + ".html#"
                    + javaDocHref.substring(1+javaDocHref.lastIndexOf('/'))
                    + "()";
        return javaDocHref;
    }

    private static TestStatus getTestStatus(String testName, String date, Map<String, Map<String, List<String>>> testResults)
    {
        String simpleName = date + "/" + testName.substring("org.thermostat.qa.testsuites.".length(), testName.lastIndexOf('.')) + ".log";
        for (Map.Entry<String, List<String>> test : testResults.get(date).entrySet())
        {
            if (test.getKey().endsWith(simpleName))
            {
                List<String> log = test.getValue();
                for (String line : log)
                {
                    if (line.startsWith("FAILED: "))
                    {
                        String tn = line.substring("FAILED: ".length());
                        if (tn.indexOf(' ') >= 0)
                        {
                            tn = tn.substring(0, tn.indexOf(' '));
                        }
                        if (tn.indexOf(':') >= 0)
                        {
                            tn = tn.substring(0, tn.indexOf(':'));
                        }
                        if (tn.equals(testName))
                        {
                            return TestStatus.FAILED;
                        }
                    }
                    if (line.startsWith("PASSED: "))
                    {
                        String tn = line.substring("PASSED: ".length());
                        if (tn.equals(testName))
                        {
                            return TestStatus.PASSED;
                        }
                    }
                    if (line.startsWith("IGNORED: "))
                    {
                        String tn = line.substring("IGNORED: ".length());
                        if (tn.equals(testName))
                        {
                            return TestStatus.IGNORED;
                        }
                    }
                }
            }
        }
        return TestStatus.NOT_FOUND;
    }

}
